/*TWINE telemetry extraction program
  In theory, should extract objects from TWINE (N64)*/
#include <stdio.h>
#include <stdlib.h>
#include "TWINE.h"
#include "TWINEobj.h"
#include "BATCH.h"

/*enumerate the different format types
  this is just for my convenience*/
enum{CI_4_15_SWAP      =8,
     CI_8_16           =9,
     CI_4_16_SWAP      =12,
     CI_4_16_SWAP_FLIP =14
     };

typedef struct 
      {int type;
       int swap;
       short width;
       short height;
       short index;
       unsigned long imgpos;
       unsigned long cipos;
       unsigned long imgsize;
       unsigned long cisize;
       } IMG;

/*functions*/
/*returns filesize, obviously*/
unsigned long filesize(FILE *in)
{fseek(in,0,SEEK_END);
return ftell(in);}

/*fill IMG struct*/
int FillIMG(unsigned long num,FILE *iimg,FILE *ipal,FILE *img,IMG *data)
{unsigned long x;
 short width=0,height=0;
 
 fseek(iimg,num*12,SEEK_SET);
 fread(&data->type,2,1,iimg);
 fread(&width,2,1,iimg);
 fread(&height,2,1,iimg);
#ifdef INTEL
 width=shortswap(width);
 height=shortswap(height);
#endif

 data->index=0;
 data->cipos=0;
 
 fseek(iimg,num*12+8,SEEK_SET);
 fread(&x,4,1,iimg);
#ifdef INTEL
 x=byteswap(x);
#endif
 data->imgpos=x;

 if(!width && !height)    /*they tend to put materials in without entries when the material isn't used*/
   {fseek(iimg,(num+1)*12+8,SEEK_SET);
    if(feof(iimg)) x=filesize(img)-x;
    else          /*guess width and height from next entry*/
      {fread(&x,4,1,iimg);
#ifdef INTEL
       x=byteswap(x);
#endif
       x-=data->imgpos;
       }
    if(data->type==8 || data->type==12 || data->type==14) x*=2;
    /*fake square root, and probably better concidering the circumstances*/
    while((width*height)<=x)
         {width++;
          height++;
          }
     width--;
     height--;
    }
 width++;
 height++;
 data->width=width;
 data->height=height;

 /*palette data*/
 fseek(ipal,num*8,SEEK_SET);
 data->index=fgetc(ipal)+1;
 fread(&x,3,1,ipal);       /*unused data*/
 fread(&x,4,1,ipal);
#ifdef INTEL
 x=byteswap(x);
#endif
 data->cipos=x*2;
 
 /*determine filesizes*/
 switch(data->type)
       {case 9:    /*8bit ci*/
             data->cisize=data->index * 2;
             data->imgsize=width*height;
             data->swap=0;
             break;
        default:   /*4bit ci - 8, 12, 14*/
             data->cisize=data->index * 2;
             data->imgsize=width*height/2;
             data->swap=1;
       }
  
 return 0;}

/*combines the palette and image*/
unsigned long Combine(FILE *out,FILE *img,FILE *pal,IMG data)
{unsigned long x;
 
 /*write palette - extend to 256 if need be*/
 fseek(pal,data.cipos,SEEK_SET);
 for(x=0;x<data.cisize;x++) fputc(fgetc(pal),out);
 while(x<data.index)
      {fputc(0,out);
       x++;
       }
 /*copy image*/
 fseek(img,data.imgpos,SEEK_SET);
 for(x=0;x<data.imgsize;x++) fputc(fgetc(img),out);
 
 return ftell(out);}

/*converts type values into something the TGA thingy can use
  0 -32bit rgba
  1 -24bit rgb
  2 -16/15bit rgba
  3 -16ci  32bit samples
  4 -16ci  24bit samples
  5 -16ci  16bit samples
  6 -256ci 32bit samples
  7 -256ci 24bit samples
  8 -256ci 16bit samples
  9 -16bit IA (8:8)
  A -8bit IA  (4:4)
  B -4bit IA  (3:1)
  C -8bit I
  D -4bit I
  0x40    flip horizontal
  0x80    flip vertical
  */
int Typify(int org)
{switch(org)
       {case 8:  return 5;
        case 9:  return 8;
        case 12: return 0x85;
        case 14: return 0xC5;
        default: return 5;
        }
 return org;}

/*main
  open files and whatnot*/
int main(int argc, char *argv[])
{char filename[130];
FILE *iimg,*img,*ipal,*pal,*mtl,*bat,*size,*out;
unsigned long x,y,file,num=0,mode=0;
unsigned short z;
IMG data;

/*open input and output files*/
if(argc<3) erratta(50);
file=strtol(argv[2],0,16);
sprintf(filename,"%s.img",argv[2]);
if(!(img=fopen(filename,"rb"))) erratta(5);
sprintf(filename,"%s.pal",argv[2]);
if(!(pal=fopen(filename,"rb"))) erratta(6);
sprintf(filename,"%s.iimg",argv[2]);
if(!(iimg=fopen(filename,"rb"))) erratta(7);
sprintf(filename,"%s.ipal",argv[2]);
if(!(ipal=fopen(filename,"rb"))) erratta(8);

sprintf(filename,"%s.mtl",argv[1]);
if(!(mtl = fopen(filename,"wt"))) erratta(9);
sprintf(filename,"%s.bat",argv[1]);
if(!(bat = fopen(filename,"wt"))) erratta(9);
sprintf(filename,"%s.bin",argv[1]);
if(!(size = fopen(filename,"wb"))) erratta(9);

if(argc>3) mode=strtol(argv[3],0,0);
if(mode>1) mode=0;

/*prep output file*/
Rem(mtl,TWINE Telemetry Extraction Program - Material File);
Rem(mtl,c/o Zoinkity (nefariousdogooder@yahoo.com));
ECHO(bat,0);
VERIFY(bat,NibbleSwapper.exe,noprog);
if(mode) VERIFY(bat,N64toTGA.exe,noTGA);
else     VERIFY(bat,N64toDIB.exe,noDIB);

num=filesize(iimg);
num/=12;
for(x=0;x<num;x++)
   {NewMtl(mtl,x);
    ImgDif(mtl,file,x,mode ? "tga":"bmp");
    
    FillIMG(x,iimg,ipal,img,&data);
    /*write size entries to correspondning .bin*/
    fseek(size,x*4,SEEK_SET);
    z=data.width*32;
    fwrite(&z,2,1,size);
    z=data.height*32;
    fwrite(&z,2,1,size);
    /*open output binary file and fill with... stuff*/
    sprintf(filename,"%X_%i.bin",file,x);
    if(!(out=fopen(filename,"wb"))) continue;
    if(data.index>16) data.index=256;      /*forces 256 color palette for output*/
    y=Combine(out,img,pal,data);
    /*determine if it needs nibbleswapping*/
    fprintf(bat,"IF EXIST %s (\n",filename);
    if(data.swap) fprintf(bat,"\tSTART /W NibbleSwapper.exe %s %#x\n",filename,y-data.imgsize);
    /*write the happy processing command*/
    fprintf(bat,"\tSTART /W N64to%s.exe %s %i %s.bin %i\n",mode ? "TGA":"DIB",filename,Typify(data.type),argv[1],x);
    fprintf(bat,"\tIF errorlevel 0 IF NOT errorlevel 1 DEL %s\n)\n",filename);
    fclose(out);
   }

/*complete the batch file*/
fprintf(bat,"Goto end\n");
ERROR(bat,noprog,Unable to find NibbleSwapper.exe\nPlace it within the working directory);
if(mode) ERROR(bat,noTGA,Unable to find N64toTGA.exe\nPlace it within the working directory);
else     ERROR(bat,noDIB,Unable to find N64toDIB.exe\nPlace it within the working directory);
fprintf(bat,"\n:end\n");

/*wrap things up*/
fclose(size);
fclose(mtl);
fclose(bat);
fclose(img);
fclose(pal);
fclose(iimg);
fclose(ipal);

sprintf(filename,"CALL %s.bat",argv[1]);
system(filename);
return 0;}
