/*This is from some old source I had lying around
  It converts N64 images to BMP format,
    usually by screwing over the original formatting.
    
    accepts:
req        filename.bin    name of N64 binary.  palettes affixed to start
req        type value      value specifying the type of image being converted
opt        filename.bin    name of size file
           index           index in size file to give the proportions of the image
*/

#include <stdio.h>
#include <stdlib.h>
#include "N64toDIB.h"
#include "TWINE.h"

/*function declarations*/
unsigned long filesize(FILE *in);
unsigned long filecopy(FILE* in,FILE* out,unsigned long offset);
unsigned long ci16to32(RGB16 org,RGBQUAD *out);
unsigned long Greyscale(unsigned char grey, unsigned char alpha,RGBQUAD *out);

int InitBMPfileHeader(BITMAPFILEHEADER *in,unsigned long imgsize, unsigned long imgoff);
int InitBMPInfoHeader(BITMAPINFOHEADER *in,long w,long h,short bits,long ci,int mode);
int InitBITMASKS(BITMASKS *in,int bits,int mode);
unsigned long WriteHeader(FILE *out,BITMAPFILEHEADER *file,BITMAPINFOHEADER *info,BITMASKS *mask,int mode);
long CItest(FILE *in, unsigned long offset,int type);

/*palette conversion macros*/
#define IA16to32(value,out) Greyscale((char)((value>>8)&0xFF),(char)(value&0xFF),out)
#define IA8to32(value,out)  Greyscale((value&0xF0) | ((value&0xF0)>>4),(value&0xF) | ((value&0xF)<<4),out)
#define IA4to32(value,out)  Greyscale(((value&0xE)<<4) | ((value&0xE)<<1) | ((value&0xE)>>2),(value&1)*0xFF,out)
#define I8to32(value,out)   Greyscale(value,0xFF,out)
#define I4to32(value,out)   Greyscale((value&0xF) | ((value&0xF)<<4),0xFF,out)


/*load some files*/
/*converts type values into something the BMP thingy can use
  0 -32bit rgba
  1 -24bit rgb
  2 -16/15bit rgba
  3 -16ci  32bit samples
  4 -16ci  24bit samples
  5 -16ci  16bit samples
  6 -256ci 32bit samples
  7 -256ci 24bit samples
  8 -256ci 16bit samples
  9 -16bit IA (8:8)
  A -8bit IA  (4:4)
  B -4bit IA  (3:1)
  C -8bit I
  D -4bit I
  0x40    flip horizontal   -    not currently implemented
  0x80    flip vertical     -    for win98+, negates height
  
  output formats
  0xx    normal bmp file
  1xx    RLE-8          - not implemented, and no plans to do so
  2xx    RLE-4          - not implemented, and no plans to do so
  3xx    bitmapX format - uses masks found between info header and palette
  8xx    DIB format     - haven't implemented properly yet
  16bit format on output
  0xxx   1:5:5:5
  1xxx   1:5:5:5 - alpha
  2xxx   5:6:5
  4xxx   5:5:5:1 - alpha  (N64)
  */
int main(int argc, char *argv[])
{FILE *in, *out, *sizes;
char filename[130];
unsigned long size,x,y,mode;
long w=0,h=0;
BITMAPFILEHEADER file;
BITMAPINFOHEADER info;
BITMASKS mask;

if(argc<3) erratta(75);
if(!(in = fopen(argv[1], "rb"))) erratta(9);
if(sizes = fopen(argv[3], "rb"))
  {y=strtol(argv[4],0,0);
   fseek(sizes,y*4,SEEK_SET);
   fread(&w,2,1,sizes);
   fread(&h,2,1,sizes);
   w/=32;
   h/=32;
   fclose(sizes);
   }

mode=strtol(argv[2],0,0);
strcpy(filename,argv[1]);
strcpy(strstr(filename,".bin"),".bmp");
if (!(out = fopen(filename, "wb+"))) erratta(1);

y=mode&0xF;
size=filesize(in);
if(mode&0x80) h=0-h;
/*if(mode&0x40) x^=0x10;*/
mode=mode>>8;

/*do stuff specific to each image type*/
if(y==0){                  /*32bit color*/
         if((mode&3)==MASKS) InitBITMASKS(&mask,32,mode);
         InitBMPInfoHeader(&info,w,h,RGBA,0,mode);
         x=14;       /*sizeof(BITMAPFILEHEADER);*/
         if(!(mode&8)) x+=info.BlockSize;
         InitBMPfileHeader(&file,info.SizeImage,x);
         WriteHeader(out,&file,&info,&mask,mode);
         /*write image*/
         for(x=0;x<size;x+=4)
            {RGBQUAD buf;
             
             fseek(in,x,SEEK_SET);
             fread(&buf,4,1,in);
             fseek(out,0,SEEK_END);
             fwrite(&buf,4,1,out);
             }
         }                 /*end 32bit color*/
else if(y==1){             /*24bit color*/
         if((mode&3)==MASKS) InitBITMASKS(&mask,24,mode);
         InitBMPInfoHeader(&info,w,h,RGB,0,mode);
         x=14;       /*sizeof(BITMAPFILEHEADER);*/
         if(!(mode&8)) x+=info.BlockSize;
         InitBMPfileHeader(&file,info.SizeImage,x);
         WriteHeader(out,&file,&info,&mask,mode);
         /*write image*/
         for(x=0;x<size;x+=3)
            {RGBQUAD buf;
             
             fseek(in,x,SEEK_SET);
             fread(&buf,3,1,in);
             fseek(out,0,SEEK_END);
             fwrite(&buf,3,1,out);
             }
         }                 /*end 24bit color*/
else if(y==2){             /*16bit color*/
         if((mode&3)==MASKS) InitBITMASKS(&mask,16,mode);
         InitBMPInfoHeader(&info,w,h,RGBA16,0,mode);
         x=14;       /*sizeof(BITMAPFILEHEADER);*/
         if(!(mode&8)) x+=info.BlockSize;
         InitBMPfileHeader(&file,info.SizeImage,x);
         WriteHeader(out,&file,&info,&mask,mode);
         /*work out which 16bit mode you're outputting into*/
         /*write image*/
         for(x=0;x<size;x+=2)
            {RGB16 buf;
             
             fseek(in,x,SEEK_SET);
             fread(&buf,2,1,in);
#ifdef INTEL
             buf.s=shortswap(buf.s);
#endif
             buf.s=pal16conv(buf,(mode&0xF0) | CI_N64);
             fseek(out,0,SEEK_END);
             fwrite(&buf,2,1,out);
             }
         }                 /*end 16bit color*/
else if(y==3){             /*16ci 32bit color*/
         if((mode&3)==MASKS) InitBITMASKS(&mask,32,mode);
         InitBMPInfoHeader(&info,w,h,CI16,CItest(in,0x40,CI16),mode);
         x=14;       /*sizeof(BITMAPFILEHEADER);*/
         x+=(info.CiNum*4);
         if(!(mode&8)) x+=info.BlockSize;
         InitBMPfileHeader(&file,info.SizeImage,x);
         WriteHeader(out,&file,&info,&mask,mode);
         /*copy palette onto that*/
         for(x=0;x<info.CiNum;x++)
            {RGBQUAD buf;
             
             fseek(in,x*4,SEEK_SET);
             fread(&buf,4,1,in);
             fseek(out,0,SEEK_END);
             fwrite(&buf,4,1,out);
             }
         /*now write the index, which shouldn't require formatting*/
         filecopy(in,out,0x40);
         }                 /*end 16ci 32bit color*/
else if(y==4){             /*16ci 24bit color*/
         if((mode&3)==MASKS) InitBITMASKS(&mask,24,mode);
         InitBMPInfoHeader(&info,w,h,CI16,CItest(in,0x30,CI16),mode);
         x=14;       /*sizeof(BITMAPFILEHEADER);*/
         x+=(info.CiNum*4);
         if(!(mode&8)) x+=info.BlockSize;
         InitBMPfileHeader(&file,info.SizeImage,x);
         WriteHeader(out,&file,&info,&mask,mode);
         /*copy palette onto that*/
         for(x=0;x<info.CiNum;x++)
            {RGBQUAD buf;
             
             fseek(in,x*3,SEEK_SET);
             fread(&buf,3,1,in);
             buf.alpha=0xFF;
             fseek(out,0,SEEK_END);
             fwrite(&buf,4,1,out);
             }
         /*now write the index, which shouldn't require formatting*/
         filecopy(in,out,0x30);
         }                 /*end 16ci 24bit color*/
else if(y==5){             /*16ci 16bit color*/
         if((mode&3)==MASKS) InitBITMASKS(&mask,16,mode);
         InitBMPInfoHeader(&info,w,h,CI16,CItest(in,0x20,CI16),mode);
         x=14;       /*sizeof(BITMAPFILEHEADER);*/
         x+=(info.CiNum*4);
         if(!(mode&8)) x+=info.BlockSize;
         InitBMPfileHeader(&file,info.SizeImage,x);
         WriteHeader(out,&file,&info,&mask,mode);
         /*copy palette onto that*/
         for(x=0;x<info.CiNum;x++)
            {long buf;
             RGB16 bufn64;
             
             fseek(in,x*2,SEEK_SET);
             fread(&bufn64,2,1,in);
#ifdef INTEL
             bufn64.s=shortswap(bufn64.s);
#endif
             buf=ci16to32(bufn64,NULL);
#ifdef INTEL
             buf=byteswap(buf);
#endif
             fseek(out,0,SEEK_END);
             fwrite(&buf,4,1,out);
             }
         /*now write the index, which shouldn't require formatting*/
         filecopy(in,out,0x20);
         }                 /*end 16ci 16bit color*/
else if(y==6){             /*256ci 32bit color*/
         if((mode&3)==MASKS) InitBITMASKS(&mask,32,mode);
         InitBMPInfoHeader(&info,w,h,CI256,CItest(in,0x400,CI256),mode);
         x=14;       /*sizeof(BITMAPFILEHEADER);*/
         x+=(info.CiNum*4);
         if(!(mode&8)) x+=info.BlockSize;
         InitBMPfileHeader(&file,info.SizeImage,x);
         WriteHeader(out,&file,&info,&mask,mode);
         /*copy palette onto that*/
         for(x=0;x<info.CiNum;x++)
            {RGBQUAD buf;
             
             fseek(in,x*4,SEEK_SET);
             fread(&buf,4,1,in);
             fseek(out,0,SEEK_END);
             fwrite(&buf,4,1,out);
             }
         /*now write the index, which shouldn't require formatting*/
         filecopy(in,out,0x400);
         }                 /*end 256ci 32bit color*/
else if(y==7){             /*256ci 24bit color*/
         if((mode&3)==MASKS) InitBITMASKS(&mask,24,mode);
         InitBMPInfoHeader(&info,w,h,CI256,CItest(in,0x300,CI256),mode);
         x=14;       /*sizeof(BITMAPFILEHEADER);*/
         x+=(info.CiNum*4);
         if(!(mode&8)) x+=info.BlockSize;
         InitBMPfileHeader(&file,info.SizeImage,x);
         WriteHeader(out,&file,&info,&mask,mode);
         /*copy palette onto that*/
         for(x=0;x<info.CiNum;x++)
            {RGBQUAD buf;
             
             fseek(in,x*3,SEEK_SET);
             fread(&buf,3,1,in);
             buf.alpha=0xFF;
             fseek(out,0,SEEK_END);
             fwrite(&buf,4,1,out);
             }
         /*now write the index, which shouldn't require formatting*/
         filecopy(in,out,0x300);
         }                 /*end 256ci 24bit color*/
else if(y==8){             /*256ci 16bit color*/
         if((mode&3)==MASKS) InitBITMASKS(&mask,16,mode);
         InitBMPInfoHeader(&info,w,h,CI256,CItest(in,0x200,CI256),mode);
         x=14;       /*sizeof(BITMAPFILEHEADER);*/
         x+=(info.CiNum*4);
         if(!(mode&8)) x+=info.BlockSize;
         InitBMPfileHeader(&file,info.SizeImage,x);
         WriteHeader(out,&file,&info,&mask,mode);
         /*copy palette onto that*/
         for(x=0;x<info.CiNum;x++)
            {long buf;
             RGB16 bufn64;
             
             fseek(in,x*2,SEEK_SET);
             fread(&bufn64,2,1,in);
#ifdef INTEL
             bufn64.s=shortswap(bufn64.s);
#endif
             buf=ci16to32(bufn64,NULL);
#ifdef INTEL
             buf=byteswap(buf);
#endif
             fseek(out,0,SEEK_END);
             fwrite(&buf,4,1,out);
             }
         /*now write the index, which shouldn't require formatting*/
         filecopy(in,out,0x200);
         }                 /*end 256ci 16bit color*/
else if(y==9){             /*16bit IA*/
         if((mode&3)==MASKS) InitBITMASKS(&mask,32,mode);
         InitBMPInfoHeader(&info,w,h,RGBA,0,mode);
         x=14;       /*sizeof(BITMAPFILEHEADER);*/
         if(!(mode&8)) x+=info.BlockSize;
         InitBMPfileHeader(&file,info.SizeImage,x);
         WriteHeader(out,&file,&info,&mask,mode);
         /*write image*/
         for(x=0;x<size;x+=2)
            {RGBQUAD buf;
             unsigned short val;
             
             fseek(in,x,SEEK_SET);
             fread(&val,2,1,in);
#ifdef INTEL
             val=shortswap(val);
#endif
             IA16to32(val,&buf);
             fseek(out,0,SEEK_END);
             fwrite(&buf,4,1,out);
             }
         }                 /*end 16bit IA*/
else if(y==10){            /*8bit IA   ccccaaaa*/
         if((mode&3)==MASKS) InitBITMASKS(&mask,32,mode);
         InitBMPInfoHeader(&info,w,h,RGBA,0,mode);
         x=14;       /*sizeof(BITMAPFILEHEADER);*/
         if(!(mode&8)) x+=info.BlockSize;
         InitBMPfileHeader(&file,info.SizeImage,x);
         WriteHeader(out,&file,&info,&mask,mode);
         /*write image*/
         for(x=0;x<size;x++)
            {RGBQUAD buf;
             unsigned char val;
             
             fseek(in,x,SEEK_SET);
             val=fgetc(in);
             IA8to32(val,&buf);
             fseek(out,0,SEEK_END);
             fwrite(&buf,4,1,out);
             }
         }                 /*end 8bit IA*/ 
else if(y==11){            /*4bit IA.  ccca*/
         if((mode&3)==MASKS) InitBITMASKS(&mask,32,mode);
         InitBMPInfoHeader(&info,w,h,RGBA,0,mode);
         x=14;       /*sizeof(BITMAPFILEHEADER);*/
         if(!(mode&8)) x+=info.BlockSize;
         InitBMPfileHeader(&file,info.SizeImage,x);
         WriteHeader(out,&file,&info,&mask,mode);
         /*write image*/
         for(x=0;x<size;x++)
            {RGBQUAD buf;
             unsigned char val;
             
             fseek(in,x,SEEK_SET);
             val=fgetc(in);
             IA4to32(val>>4,&buf);
             fseek(out,0,SEEK_END);
             fwrite(&buf,4,1,out);
             /*other nibble*/
             IA4to32(val&0xF,&buf);
             fseek(out,0,SEEK_END);
             fwrite(&buf,4,1,out);
             }
         }                 /*end 4bit IA*/
else if(y==12){            /*8bit I*/
         if((mode&3)==MASKS) InitBITMASKS(&mask,24,mode);
         InitBMPInfoHeader(&info,w,h,RGB,0,mode);
         x=14;       /*sizeof(BITMAPFILEHEADER);*/
         if(!(mode&8)) x+=info.BlockSize;
         InitBMPfileHeader(&file,info.SizeImage,x);
         WriteHeader(out,&file,&info,&mask,mode);
         /*write image*/
         for(x=0;x<size;x++)
            {RGBQUAD buf;
             unsigned char val;
             
             fseek(in,x,SEEK_SET);
             val=fgetc(in);
             I8to32(val,&buf);
             fseek(out,0,SEEK_END);
             fwrite(&buf,3,1,out);
             }
         }                 /*end 8bit I*/
else if(y==13){            /*4bit I*/
         if((mode&3)==MASKS) InitBITMASKS(&mask,24,mode);
         InitBMPInfoHeader(&info,w,h,RGB,0,mode);
         x=14;       /*sizeof(BITMAPFILEHEADER);*/
         if(!(mode&8)) x+=info.BlockSize;
         InitBMPfileHeader(&file,info.SizeImage,x);
         WriteHeader(out,&file,&info,&mask,mode);
         /*write image*/
         for(x=0;x<size;x++)
            {RGBQUAD buf;
             unsigned char val;
             
             fseek(in,x,SEEK_SET);
             val=fgetc(in);
             I4to32(val>>4,&buf);
             fseek(out,0,SEEK_END);
             fwrite(&buf,3,1,out);
             /*do other nibble*/
             I4to32(val&0xF,&buf);
             fseek(out,0,SEEK_END);
             fwrite(&buf,3,1,out);
             }
         }                 /*end 4bit I*/
else {fclose(out);
      fclose(in);
      remove(filename);
      return 1;}
    
fclose(in);
fclose(out);
return 0;
}

/*****************************************
converts 8bit greyscale into 32bit white
alpha optional
*****************************************/
unsigned long Greyscale(unsigned char grey, unsigned char alpha,RGBQUAD *out)
{RGBQUAD buf;
 buf.red=grey;
 buf.green=grey;
 buf.blue=grey;
 buf.alpha=alpha;
 if(out!=NULL) *out=buf;
 return ((grey<<24)|(grey<<16)|(grey<<8)|alpha);}

/*****************************************
palette16() converts the common 16 color
palettes in N64 games into RGB values,
mostly to insert into BMP images that
are later compressed into PNGs

Anyway, feed it a 16bit palette and you 
get RGB.  Loop to get more than one
******************************************/
unsigned long ci16to32(RGB16 org,RGBQUAD *out)
{RGBQUAD buf;
 buf.red=org.n.red<<3;
 buf.green=org.n.green<<3;
 buf.blue=org.n.blue<<3;
 buf.alpha=org.n.alpha * 0xFF;
 /*this allows you to pass a NULL pointer*/
 if(out!=NULL) *out=buf;
 /*returns RGBA value as a long*/
 return ((buf.red<<24)|(buf.green<<16)|(buf.blue<<8)|(buf.alpha));
}    

/****************************************/
/*get the size of a file
it happens often enough it might as well be a function
plus, it won't mess with CUR_POS*/
/****************************************/
unsigned long filesize(FILE *in)
{fseek(in,0,SEEK_END);
return ftell(in);}

/****************************************/
/*appends contents of file in to out*/
/****************************************/
unsigned long filecopy(FILE* in,FILE* out,unsigned long offset)
{unsigned long x,y;

x=filesize(in);
fseek(in,offset,SEEK_SET);
fseek(out,0,SEEK_END);
for(y=0;y<x;y++) fputc(fgetc(in),out);
return y;}            

/****************************************
 sets data into the file's (DIB) header
 ****************************************/
int InitBMPfileHeader(BITMAPFILEHEADER *in,unsigned long imgsize, unsigned long imgoff)
{in->Type=0x4D42;
 in->Reserved1=0;
 in->Reserved2=0;
 in->Size=imgsize+imgoff;
 in->ImgOffset=imgoff;
 
 return 0;}

/****************************************
 sets data into the info header
 type is roughly equivalent to DataFormat
 ****************************************/
int InitBMPInfoHeader(BITMAPINFOHEADER *in,long w,long h,short bits,long ci,int mode)
{in->Height=h;
 in->Width=w;
 in->CiNum=ci;
 in->CiMin=ci;       /*every last pixel is of the utmost importance! (apparently...)*/
 in->Bits=bits;
 
 in->Planes=1;
 in->XPelsPerMeter=2835;
 in->YPelsPerMeter=2835;
 
 in->DataFormat=mode&7;
 if((mode&3)==MASKS) in->BlockSize=0x38;
 else in->BlockSize=0x28;
 /*compute size from proportions and bitdepth*/
 bits/=4;
 in->SizeImage=(w*h*bits)/2;
 
 return 0;}

/****************************************
 writes bitmasks for appropriate format
 mode designates the 16bit formats
 everything else is quite standard
 ****************************************/
int InitBITMASKS(BITMASKS *in,int bits,int mode)
{mode=(mode>>4)&7;
 in->red=0;
 in->green=0;
 in->blue=0;
 in->alpha=0;
 
 if(bits>=24)
   {in->red=0xFF000000;
    in->green=0x00FF0000;
    in->blue=0x0000FF00;
    }
 if(bits==32) in->alpha=0x000000FF;
 if(bits==16)
   {if(mode==1) in->alpha=0x8000;
    if(mode<2)         /*1:5:5:5 - alpha1, flat0*/
      {in->red=0x7C00;
       in->green=0x3E0;
       in->blue=0x1F;
       }
    if(mode==2)        /*5:6:5*/
      {in->red=0xF800;
       in->green=0x7E0;
       in->blue=0x1F;
       }
    if(mode==4)        /*N64 format*/
      {in->red=0xF800;
       in->green=0x7C0;
       in->blue=0x3E;
       in->alpha=0x1;
       }
    }
 
 return 0;}
 
/****************************************
 writes the two info blocks, as well as bitmask when the mode supports it
 ****************************************/
unsigned long WriteHeader(FILE *out,BITMAPFILEHEADER *file,BITMAPINFOHEADER *info,BITMASKS *mask,int mode)
{unsigned long x;

 fseek(out,0,SEEK_SET);
 /*write DIB header*/
 x=14;   /*sizeof(BITMAPFILEHEADER);*/
 fwrite(&file->Type,2,1,out);         /*slight problem with it padding my struct*/
 fwrite(&file->Size,4,1,out);
 fwrite(&file->Reserved1,2,1,out);
 fwrite(&file->Reserved2,2,1,out);
 fwrite(&file->ImgOffset,4,1,out);
 
 if(mode&8) return ftell(out);
 /*write BMP header*/
 x=sizeof(BITMAPINFOHEADER);
 fwrite(info,x,1,out);
 /*write masks if applicable*/
 x=sizeof(BITMASKS);
 if(mode&MASKS) fwrite(mask,x,1,out);
 
 return ftell(out);}

/****************************************
 Horrible thingy to determine 
 actual max palette index used
 ****************************************/
long CItest(FILE *in, unsigned long offset,int type)
{long x=0;
 unsigned long y,z,size;
 
 size=filesize(in);
 for(y=offset;y<size;y++)
    {fseek(in,y,SEEK_SET);
     z=fgetc(in);
     if(type==CI16)
       {if((z&0xF)>x) x=(z&0xF);
        z&=0xF0;
        z=z>>4;
        if(x>15) return -1;
        }
     if(z>x) x=z;
     }
 if(x>255) return -1;
 
 return x+1;}
