#ifndef N64_TO_DIB
#define N64_TO_DIB

#include <stdio.h>  /*stupid compiler being ansy*/
#include <stdlib.h>

/*typedefs and global variables*/
typedef struct {    /* bmfh */
    unsigned short    Type;
    unsigned long     Size;
    unsigned short    Reserved1;
    unsigned short    Reserved2;
    unsigned long     ImgOffset;
} BITMAPFILEHEADER;

typedef struct {    /* bmih */
    unsigned long    BlockSize;               /*0x28, unless format 3*/
    long             Width;
    long             Height;                  /*inverts if negative - win98+*/
    short            Planes;                  /*always 1?*/
    short            Bits;
    long             DataFormat;              /*0-uncompressed, 1-RLE8, 2-RLE4, 3-declared RGBA masks*/
    unsigned long    SizeImage;
    long             XPelsPerMeter;           /*horizontal resolution*/
    long             YPelsPerMeter;           /*vertical resolution*/
    long             CiNum;                   /*#colors in palette*/
    long             CiMin;                   /*#colors crappy monitors should use minimum*/
} BITMAPINFOHEADER;

typedef struct {
    unsigned long    red;
    unsigned long    green;
    unsigned long    blue;
    unsigned long    alpha;
} BITMASKS;

/*DataFormat values*/
#define UNCOMPRESSED  0
#define RLE8          1
#define RLE4          2
#define MASKS         3   /*declares RGBA masks to apply to each sample*/

/*Bits - bitsamples for different data types*/
#define MONOCHROME   1
#define CI16         4
#define CI256        8
#define RGBA16       16
#define RGB          24
#define RGBA         32

typedef struct {
    unsigned char blue;
    unsigned char green;
    unsigned char red;
    unsigned char alpha;
} RGBQUAD;

typedef struct {
    unsigned int red    :5;
    unsigned int green  :5;
    unsigned int blue   :5;
    unsigned int alpha  :1;
} RGB1555;

typedef struct {
    unsigned int alpha  :1;
    unsigned int red    :5;
    unsigned int green  :5;
    unsigned int blue   :5;
} RGBN64;

typedef struct {
    unsigned int red    :5;
    unsigned int green  :6;
    unsigned int blue   :5;
} RGB565;

typedef union {
    unsigned short s;
    RGB1555        a;
    RGB565         g;
    RGBN64         n;
} RGB16;

/*takes N64 16bit color and shoves it into one of several structs
  0x  -:5:5:5 out
  1x  1:5:5:5 out
  2x  5:6:5 out
  4x  5:5:5:1 (N64) out
  x0  -:5:5:5 in
  x1  1:5:5:5 in
  x2  5:6:5 in
  x4  5:5:5:1 (N64) in
  */
#define CI_16    0x00
#define CI_A     0x01
#define CI_G     0x02
#define CI_N64   0x04

unsigned short pal16conv(RGB16 in,int mode)
{RGB16 out;
 
 if((mode>>4)==(mode&0xF)) return in.s;
 out.s=0;
 
 if(mode&CI_G)       /*CI_G in*/
   {mode=mode>>4;
    if(mode&CI_N64)
      {out.n.red=  in.g.red;
       out.n.green=in.g.green>>1;
       out.n.blue= in.g.blue;
       out.n.alpha=1;
       }
    else
      {out.a.red=  in.g.red;
       out.a.green=in.g.green>>1;
       out.a.blue= in.g.blue;
       out.a.alpha=1;
       }
    }
 else if(mode&CI_N64)/*N64 in*/
   {mode=mode>>4;
    if(mode&CI_G)
      {out.g.red=  in.n.red;
       out.g.green=in.n.green<<1;
       out.g.blue= in.n.blue;
       }
    else 
      {out.a.red=  in.n.red;
       out.a.green=in.n.green;
       out.a.blue= in.n.blue;
       out.a.alpha=in.n.alpha;
       }
    }
 else                /*CI_A and CI_16 in*/
   {mode=mode>>4;
    if(mode&CI_G)
      {out.g.red=  in.a.red;
       out.g.green=in.a.green<<1;
       out.g.blue= in.a.blue;
       }
    else if(mode&CI_N64)
      {out.n.red=  in.a.red;
       out.n.green=in.a.green;
       out.n.blue= in.a.blue;
       out.n.alpha=in.a.alpha;
       }
    }
 
 
 return out.s;}


#endif  /*N64_TO_DIB*/
