/*TWINE image retrieval service
  sets up a batch file to extract images from TWINE (N64)*/
#define INTEL       /*define this on intel systems so values are byteswapped*/

#include <stdio.h>
#include <stdlib.h>
#include "TWINE.h"
#include "BATCH.h"

/*macros*/


/*grab image attributes
  return 0 if the files already exist, or base for data otherwise*/
unsigned long GrabIMGResources(FILE *rom,unsigned long num,FILE *bat)
{unsigned long x,y,z,pos;
 unsigned char filename[16],buf[4];
 FILE *data;
 
 /*get base image data address*/
 fseek(rom,ROMTABLE+num*16,SEEK_SET);
 fread(&pos,4,1,rom);
#ifdef INTEL
 pos=byteswap(pos);
#endif
 
 /*get the four filenames*/
 for(x=16;x<48;x+=8)
    {fseek(rom,pos+CORRECT+x,SEEK_SET);
     fread(&y,4,1,rom);
     fread(&z,4,1,rom);
#ifdef INTEL
     y=byteswap(y);
     z=byteswap(z);
#endif
     if(y==0) continue;
     
     switch(x)
           {case 16: sprintf(filename,"%X.ipal",num);
                 z*=8;
                 break;
            case 24: sprintf(filename,"%X.iimg",num);
                 z*=12;
                 break;
            case 32: sprintf(filename,"%X.pal",num);
                 break;
            case 40: sprintf(filename,"%X.img",num);
                 break;
            }
     /*only write files when they don't exist*/
     if(data=fopen(filename,"rb")) 
       {fclose(data);
       return 0;
       }
     if(!(data=fopen(filename,"wb+"))) erratta(1);
     /*determine if it is EDL compressed as well*/
     y+=CORRECT;
     fseek(rom,y,SEEK_SET);
     fread(buf,1,4,rom);
     if(!strncmp(buf,"EDL",3))
       {fread(&z,4,1,rom);
#ifdef INTEL
        z=byteswap(z);
#endif
        DECOMP(bat,filename);
        }
     /*grab the actual file*/
     for(fseek(rom,y,SEEK_SET);z>0;z--) fputc(fgetc(rom),data);
     fclose(data);
     }
 
 return pos;}


/*main
  open files and whatnot*/
int main(int argc, char *argv[])
{char filename[130];
FILE *out,*obj,*rom;
unsigned long x,y,loc=0,num,flag=2;

for(x=1;x<argc;x++)/*command-line arguments*/
    {if(argv[x][0]=='-' || argv[x][0]=='/')
      {switch(argv[x][1])
             {case 'C': case 'c': flag^=2; break;
              case 'T': case 't': flag|=1; break;
              case 'B': case 'b': flag&=0xFFFFFFFE; break;
              case 'H': case 'h': case '?':
                  printf("\nTWINE (N64) image extraction program");
                  printf("\nusage: %s <filename.rom>",basename(argv[0]));
                  printf("\n+Requires the file 'objects' already be extracted!");
                  printf("\n  /C or -C\tdo not clean extracted files");
                  printf("\n  /B or -B\toutput bitmap (.bmp) images");
                  printf("\n  /T or -T\toutput targa (.tga) images");
                  printf("\n  /H, -H, or -?\tdisplay this help message");
                  printf("\n\n\tPress -enter- to quit\n"); getchar(); return 0;
              }    
      }
    else loc=x;    /*keep track of which one is the filename*/
    }/*end command-line arguments*/

if(!loc) erratta(5);
if(!(rom=fopen(argv[loc],"rb")))    erratta(14);
if(!(obj=fopen("objects","rb")))    erratta(3);
if(!(out=fopen("images.bat","wt"))) erratta(1);

/*determine the number of images declared in the index*/
fseek(obj,0,SEEK_END);
num=ftell(obj)/0x44;

/*initialize the batch file*/
fprintf(out,"@Echo Off\n");
VERIFY(out,EDLdec.exe,baddec);
VERIFY(out,IMGFormat.exe,noprog);

/*loop extraction for each image*/
for(x=0;x<num;x++)
  {/*determine what files must be sucked out*/
  fseek(obj,x*68 + 40,SEEK_SET);
  fread(&loc,4,1,obj);
#ifdef INTEL
  loc=byteswap(loc);
#endif
  loc-=0x800C2740;
  loc/=2;
  /*get the image resources - returns data address*/
  y=GrabIMGResources(rom,loc,out);
  if(!y) continue;
  fprintf(out,"START /W IMGFormat.exe %X %X %X\nIF errorlevel 1 GOTO nofile\nCALL %X.bat\n\n",y,loc,flag&1,y);
  }/*end loop for each image*/

/*trail off the batch file*/
fprintf(out,"GOTO fin\n");                           /*EOF*/
ERROR(out,baddec,Unable to decompress file(s).  Place EDLdec.exe in working directory.);
ERROR(out,noprog,Unable to find program(s));
ERROR(out,nofile,Failure to locate file(s));
fprintf(out,":fin\n");
if(flag&2) fprintf(out,"DEL /F *.iimg\nDEL /F *.img\nDEL /F *.ipal\nDEL /F *.pal\n");
fprintf(out,":end\nEcho All done.\n");                           /*EOF*/
fflush(out);
/*wrap things up*/
fclose(rom);
fclose(obj);
fclose(out);
return 0;}
